from __future__ import annotations
from typing import List, Union
import os
from pathlib import Path
from datetime import datetime
from struct import unpack
from dataclasses import dataclass
from collections import UserList
import numpy as np
from numpy.typing import NDArray

from .gwyddion import Gwy


@dataclass
class MulImage:
    """Class for a single STM-image and its metadata in a mulfile

    Attributes:
        filepath: Path of the .mul or .flm file
        img_num: Number of the image in the file (starting at 1)
        img_id: Filename without extension with underscore and img_num
        size:
        xres: Resolution in x-direction (in pixels)
        yres: Resolution in y-direction (in pixels)
        zres:
        datetime: Datetime of image acquisition (YYYY-MM-DD hh:mm:ss)
        xsize: Size of the image in x-direction (in nm)
        ysize: Size of the image in y-direction (in nm)
        xoffset: Offset of the image in x-direction from 0 (in nm)
        yoffset: Offset of the image in y-direction from 0 (in nm)
        zscale: ADC range (in V)
        tilt: Rotation of x-scan-direction (in degree)
        speed: Image acquisition time (in seconds)
        line_time: Time for one scan line (in milliseconds)
        bias: Bias Voltage (in mV)
        current: Tunneling Current Setpoint (in nA)
        sample: Sample name as prompted when saving the image
        title: Title name as prompted when saving the image
        postpr:
        postd1:
        mode: Scan Mode
        currfac: Current factor (either 1 or 5)
        num_pointscans: Number of pointscans during image acquisition
        unitnr:
        version:
        gain: Gain of PI
        img_data: Data of the image as 2D numpy array, ordered according to
            scanning sequence (in nm)

    """

    filepath: Union[str, Path]
    img_num: int
    img_id: str
    size: int
    xres: int
    yres: int
    zres: int
    datetime: datetime
    xsize: int
    ysize: int
    xoffset: int
    yoffset: int
    zscale: int
    tilt: int
    speed: float
    line_time: float
    bias: float
    current: float
    sample: str
    title: str
    postpr: int
    postd1: int
    mode: int
    currfac: int
    num_pointscans: int
    unitnr: int
    version: int
    gain: int
    img_data: NDArray[np.float64]

    def __add__(self, other: MulImage) -> Mul:
        """Defines the addition operator for MulImage-instances

        Args:
            other: the MulImage that is added

        Returns:
            A new instance of Mul containing self and other
        """

        return Mul([self, other])

    def save_gwy(self, output_name: str) -> None:
        """Saves a MulImage-instance as a GWY file

        Args:
            output_name: name of the .gwy file to create

        """

        gwy = Gwy(self)
        gwy.save_gwyfile(output_name)


class Mul(UserList[MulImage]):
    """Class for an entire mulfile, list-like sequence of MulImage STM-images

    Args:
        data: the data the Mul-instance contains

    """

    def __init__(self, data: List[MulImage]) -> None:
        super().__init__(self)
        self.data = data

    def save_gwy(self, output_name: str) -> None:
        """Saves a Mul-instance as a GWY file

        Args:
            output_name: name of the .gwy file to create

        """

        gwy = Gwy(self)
        gwy.save_gwyfile(output_name)


def read_mul(filepath: Union[str, Path]) -> Mul:
    """Parses a mul-file (.mul) or a flm-file (.flm)

    Args:
        filepath: absolute or relative path to the .mul or .flm file

    Returns:
        An instance of Mul if more than one image was found in the file
        or an instance of Mul if the file contains one image

    """

    basename = os.path.basename(filepath)
    filename, fileext = os.path.splitext(basename)
    if fileext not in [".mul", ".flm"]:
        raise ValueError("Filetype not supported. Select a .mul or .flm file")

    MUL_BLOCK = 128
    filesize = os.path.getsize(filepath)
    data: List[MulImage] = []

    with open(filepath, "rb") as f:
        nr = unpack("h", f.read(2))[0]
        adr = unpack("i", f.read(4))[0]
        f.seek(0)

        block_counter = 0

        if adr == 3:
            f.seek(MUL_BLOCK * adr)
            block_counter += adr

        while block_counter * MUL_BLOCK < filesize:
            img_num, size = unpack("hh", f.read(4))
            xres, yres, zres = unpack("hhh", f.read(6))
            year, month, day = unpack("hhh", f.read(6))
            hour, minute, second = unpack("hhh", f.read(6))

            xsize = unpack("h", f.read(2))[0] * 0.1  # in nm
            ysize = unpack("h", f.read(2))[0] * 0.1  # in nm
            xoffset = unpack("h", f.read(2))[0] * 0.1  # in nm
            yoffset = unpack("h", f.read(2))[0] * 0.1  # in nm
            zscale, tilt = unpack("hh", f.read(4))  # in V, deg

            speed, bias, current = unpack("hhh", f.read(6))

            sample = "".join([chr(i) for i in f.read(21)])
            title = "".join([chr(i) for i in f.read(21)])

            postpr, postd1, mode, currfac = unpack("hhhh", f.read(8))
            num_pointscans, unitnr, version = unpack("hhh", f.read(6))

            spare_48, spare_49, spare_50 = unpack("hhh", f.read(6))
            spare_51, spare_52, spare_53 = unpack("hhh", f.read(6))
            spare_54, spare_55, spare_56 = unpack("hhh", f.read(6))
            spare_57, spare_58, spare_59 = unpack("hhh", f.read(6))
            gain, spare_61, spare_62, spare_63 = unpack("hhhh", f.read(8))

            img_data_raw = np.frombuffer(f.read(xres * yres * 2), dtype=np.int16)
            img_data: NDArray[np.float64] = img_data_raw.astype("float64")
            img_data *= -0.1 / 1.36 * zscale / 2000  # in nm
            img_data = img_data.reshape((xres, yres))

            if num_pointscans > 0:
                for _ in range(0, num_pointscans):
                    ps_size, ps_type, ps_time4scan = unpack("hhh", f.read(6))
                    ps_minv, ps_maxv = unpack("hh", f.read(4))
                    ps_xpos, ps_ypos = unpack("hh", f.read(4))
                    ps_dz, ps_delay = unpack("hh", f.read(4))
                    ps_version, ps_indendelay = unpack("hh", f.read(4))
                    ps_xposend, ps_yposend = unpack("hh", f.read(4))
                    ps_vt_fw, ps_it_fw = unpack("hh", f.read(4))
                    ps_vt_bw, ps_it_bw = unpack("hh", f.read(4))
                    ps_lscan = unpack("h", f.read(2))[0]

                    f.read(MUL_BLOCK - 18 * 2)  # spare
                    ps_data = np.frombuffer(f.read(ps_size * 2), dtype=np.int16)

            block_counter += size

            speed *= 0.01  # in s
            line_time = speed / yres * 1000  # in ms
            bias = -bias / 3.2768  # in mV
            current *= currfac * 0.01  # in nA

            data.append(
                MulImage(
                    filepath,
                    img_num,
                    filename + "_" + str(img_num),
                    size,
                    xres,
                    yres,
                    zres,
                    datetime(year, month, day, hour, minute, second),
                    xsize,
                    ysize,
                    xoffset,
                    yoffset,
                    zscale,
                    tilt,
                    speed,
                    line_time,
                    bias,
                    current,
                    sample,
                    title,
                    postpr,
                    postd1,
                    mode,
                    currfac,
                    num_pointscans,
                    unitnr,
                    version,
                    gain,
                    img_data,
                )
            )

    return Mul(data)



from typing import Union
from pathlib import Path
from . import mul


def load(file: Union[str, Path]) -> mul.Mul:
    """Parses a mul-file (.mul) or a flm-file (.flm)

    Args:
        file: The .mul or .flm file to be parsed

    Returns:
        An instance of Mul if more than one image was found in the file
        or an instance of Mul if the file contains one image

    """

    return mul.read_mul(file)



from __future__ import annotations
from typing import List, Union, TYPE_CHECKING
from collections.abc import Iterable
from struct import pack
import numpy as np

if TYPE_CHECKING:
    from .mul import Mul, MulImage


class Gwy:
    """Class for handeling the construction of bytes for a .gwy-file

    Args
        imgs: the Mul or MulImage instance that is converted to bytes
             according to the GWY file format

    """

    def __init__(self, imgs: Union[Mul, MulImage]) -> None:
        self.imgs: Union[Mul, List[MulImage]]

        if isinstance(imgs, Iterable):
            self.imgs = imgs
        else:
            self.imgs = [imgs]

        self.gwycontent = self._make_gwy()

    def _make_si_unit(self, dimension: str) -> bytes:
        """Constructs GwySIUnit

        Args:
            dimension: spacial dimension of the SI Unit (either xy or z)

        Returns:
            GwySIUnit bytes

        """

        si_name = (
            b"si_unit_" + bytes(dimension, "utf-8") + b"\0o" + b"GwySIUnit\0"
        )
        unit = b"unitstr\0s" + b"m\0"
        unit_size = pack("<i", len(unit))

        return si_name + unit_size + unit

    def _make_datafield(self, img: MulImage) -> bytes:
        """Constructs a GwyDatafield as defined by the .gwy file format

        Args:
            img: the image for which the datafield bytes are created

        Returns:
            GwyDataField bytes

        """

        datafield = b"".join(
            [
                b"xreal\0d" + pack("<d", float(img.xsize) * 1e-9),
                b"yreal\0d" + pack("<d", float(img.ysize) * 1e-9),
                b"xoff\0d" + pack("<d", 0),
                b"yoff\0d" + pack("<d", 0),
                b"xres\0i" + pack("<i", img.xres),
                b"yres\0i" + pack("<i", img.yres),
            ]
        )

        datafield += self._make_si_unit("xy")
        datafield += self._make_si_unit("z")

        data_arr = np.flip(img.img_data, axis=0).flatten() * 1e-9
        data_arr_size = len(data_arr)

        datafield += b"data\0D" + pack("<i", data_arr_size)
        datafield += data_arr.tobytes()

        datafield_size = pack("<i", len(datafield))

        return b"GwyDataField\0" + datafield_size + datafield

    def _make_meta(self, img: MulImage, channel_num: int) -> bytes:
        """Constructs a  GwyContainer containing the meta-data of an image

        Args:
            img: the image for which the datafield bytes are created
            channel_num: the number of the channel in the GWY file

        Return:
            GwyContainer containing meta-data in bytes

        """

        meta_name = bytes(f"/{channel_num}/meta\0o", "utf-8")
        meta_name += b"GwyContainer\0"

        meta_data = b"".join(
            [
                bytes(f"ID\0s{img.img_id}\0", "utf-8"),
                bytes(f"Bias\0s{img.bias * 1e-3:.5f} V\0", "utf-8"),
                bytes(f"Current\0s{img.current:.2f} nA\0", "utf-8"),
                bytes(f"Current factor\0s{img.currfac}\0", "utf-8"),
                bytes(f"Scan Size\0s{img.xsize} nm, {img.ysize} nm\0", "utf-8"),
                bytes(f"Resolution\0s{img.xres}, {img.yres}\0", "utf-8"),
                bytes(f"Datetime\0s{img.datetime}\0", "utf-8"),
                bytes(f"Gain\0s{img.gain}\0", "utf-8"),
                bytes(f"Mode\0s{img.mode}\0", "utf-8"),
                bytes(f"Postprocessing\0s{img.postpr}\0", "utf-8"),
                bytes(f"Scan Duration\0s{img.speed:.2f} s\0", "utf-8"),
                bytes(f"Line Time\0s{img.line_time:.2f} ms\0", "utf-8"),
                bytes(f"Tilt\0s{img.tilt} deg\0", "utf-8"),
                bytes(f"X-Offset\0s{img.xoffset:.2f} nm\0", "utf-8"),
                bytes(f"Y-Offset\0s{img.yoffset:.2f} nm\0", "utf-8"),
                bytes(f"Z-Scale\0s{img.zscale}\0", "utf-8"),
            ]
        )

        meta_data_size = pack("<i", len(meta_data))

        return meta_name + meta_data_size + meta_data

    def _make_data_container(self) -> bytes:
        """Constructs a top-level GwyContainer containing image data as
        GwyDataFields and image meta-data as nested GwyContainers

        Returns:
            Top-level GwyConainer with DataFields and meta-data in bytes

        """

        container = b""
        for channel_num, img in enumerate(self.imgs):
            channel_title = bytes(f"{img.img_id}\0", "utf-8")
            channel_key = bytes(f"/{channel_num}/data\0o", "utf-8")
            title_key = bytes(f"/{channel_num}/data/title\0s", "utf-8")

            container += (
                channel_key
                + self._make_datafield(img)
                + title_key
                + channel_title
                + self._make_meta(img, channel_num)
            )

        container_size = pack("<i", len(container))

        return b"GwyContainer\0" + container_size + container

    def _make_gwy(self) -> bytes:
        """Constructs the content of a .gwy-file as defined by the GWY file format

        Returns:
            Bytes of the complete GWY file

        """

        GWY_HEADER = b"GWYP"
        content = GWY_HEADER + self._make_data_container()
        return content

    def save_gwyfile(self, output_name: str) -> None:
        """Writes the content of a .gwy file into a file

        Args:
            output_name: name of the .gwy file to create

        """

        with open(output_name, "wb") as f:
            f.write(self.gwycontent)
